#!/usr/bin/env python
'''
  #############################################################################
  #
  #     This Cplant(TM) source code is the property of Sandia National
  #     Laboratories.
  #
  #     This Cplant(TM) source code is copyrighted by Sandia National
  #     Laboratories.
  #
  #     The redistribution of this Cplant(TM) source code is subject to the
  #     terms of the GNU Lesser General Public License
  #     (see cit/LGPL or http://www.gnu.org/licenses/lgpl.html)
  #
  #     Cplant(TM) Copyright 1998, 1999, 2000, 2001, 2002 Sandia Corporation.
  #     Under the terms of Contract DE-AC04-94AL85000, there is a non-exclusive
  #     license for use of this work by or on behalf of the US Government.
  #     Export of this program may require a license from the United States
  #     Government.
  #
  #############################################################################

 File    : framework.py
 Revision: $Revision: 1.12 $
 Author  : William McLendon, Narayan Desai
 Description:
        implementation of test driver framework.
        Contains the following classes:
                TestList
		TestSet
'''
import re
import sys
from md5 import md5
from string import join
from time import sleep
from xml.dom.minidom import parse,Document
from graph import DiGraph
from exceptions import Exception
from errorhandlers import TestSetupError,TestError
from testbase import Node,Msgbuf,TestInfo,dbWrite
from filedata import filedata
from string import strip



# Import the new testHandler class 
# ADD CUSTOM TESTS TO gTestClassMap list {key=<element name>:<testHandler class> }
from testlibs import SSSTest,shellTest,tcpipTest,failTest,passTest
gTestClassMap = {'ssslibTest':SSSTest, 'shellTest':shellTest, 'tcpipTest':tcpipTest, \
                 'failTest':failTest, 'passTest':passTest}




# ------------------------------------------------------------------------------
# class TestList
# ------------------------------------------------------------------------------
class TestList(Node):
    ''' class TestList: high level object containing the DOM trees for all test bodies. '''
    __attrlist__=['numReps','iterDelay','preDelay','postDelay']

    testtypes=[]
    typemap=gTestClassMap

    for k in gTestClassMap.keys():
        testtypes.append(k)

    def __setdefaults__(self):
        self.options   = {}
        self.numReps   = '1'
        self.iterDelay = '0.0'
        self.preDelay  = '0.0'
        self.postDelay = '0.0'

    def getDeps(self,i, child):
        ''' returns the list of dependencies as tuples (name,type,minPctMatch,maxPctMatch)'''
        deps = []
        if child.tagname=='dependency':
          deps.append( (child.name,child.type, child.minPctMatch, child.maxPctMatch) )
        if hasattr(child, 'children'):
          for c in child.children:
            deps += self.getDeps(i+1,c)
        return deps

    def prepare(self, options, filedata):
        '''TestList.prepare()'''
        self.filedata = filedata
        self.options = options
        self.scratch = {}
        self.tests=[]

        # Sort out the tests into a proper execution order (Topological Sort)
        G = DiGraph()
        for child in self.children:
            G.add_vertex(child.name)
        for child in self.children:
            deps = self.getDeps(0,child)
            for i in deps:
              if self.options.has_key('debug'): 
                  print "D:\tAdding order dependency edge: %s->%s"%(child.name,i[0])
              G.add_edge(child.name, i[0])
        G.reverse()	# reverse the edges
        L = G.topological_sort()
        # Kludgy way to do subtree work
	doAppend = 1 
        if self.options.has_key('test') and self.options['test'] in L:
            if self.options.has_key('noAncestors'):
                doAppend=0
        for t in L:
            if self.options.has_key('test') and self.options['test']==t:
                doAppend=1
            if doAppend:
                for child in self.children:
                    if (child.tagname in self.testtypes) and (child.name == t):
                        self.tests.append(child)
            if self.options.has_key('test') and self.options['test']==t:
                if self.options.has_key('noDescendents'):
                    doAppend=0

    def findTestByName(self, tname):
        for t in self.tests:
            if t.name == tname:
                return t
        return None

    def newTag(self, tagname):
        d = {'tag':tagname}
        d['attributes'] = {}
        d['children'] = []
        return d

    def pcheck2(self, itest):
        ''' check the prerequisites '''
        status = 1
        if self.options.has_key('debug'): print 'D:\tpcheck() <%s>'%(itest.tagname)
        d = self.newTag(itest.tagname)
        if itest.tagname in ['OR','NOR','AND','NAND','NOT']:
            d['tag'] = 'join'
            d['attributes']['type'] = itest.tagname
        if itest.tagname != 'dependency':
            if hasattr(itest,'children'):
                rlist=[]
                if itest.tagname == 'NOT' and len(itest.children)>1:
                    raise TestError, '<NOT> must have only one child node'

                for i in itest.children:
                    status = 1
                    tval = self.pcheck2(i)
                    if tval['attributes'].has_key('status'):
                          d['attributes']['status'] = tval['attributes']['status']
                          if d['attributes']['status'] == 'FAIL': status = 0
                    d['children'].append(tval)
                    rlist.append(status)
                if itest.tagname=='OR' or itest.tagname=='NOR':
                    status = 0
                    for i in rlist:
                        if i == 1:
                            status = 1
                    if itest.tagname=='NOR': 
                        status = not status
                    if status==0: d['attributes']['status'] = 'FAIL'
                    else: d['attributes']['status'] = 'PASS'
                elif itest.tagname=='AND' or itest.tagname=='NAND':
                    status = 1
                    for i in rlist:
                        if i == 0:
                            status = 0
                            pass
                    if itest.tagname=='NAND': 
                        pass
                        status = not status
                    if status==0: d['attributes']['status'] = 'FAIL'
                    else: d['attributes']['status'] = 'PASS'

                elif itest.tagname=='NOT':
                    if d['attributes']['status'] == "PASS": d['attributes']['status'] = "FAIL"
                    else: d['attributes']['status'] = 'PASS'
        elif itest.tagname=='dependency':
            parent = self.findTestByName(itest.name)
            if parent:				# wcm- for kludge to allow --test, --noDescendents, etc to work
               d['attributes']['name'] = parent.name
               d['attributes']['minPctExpect'] = itest.minPctMatch
               d['attributes']['maxPctExpect'] = itest.maxPctMatch
               d['attributes']['pctActual'] = parent.pctMatched
               if parent.pctMatched < float(itest.minPctMatch) or \
                    parent.pctMatched > float(itest.maxPctMatch):
                   status = 0
                   d['attributes']['status'] = 'FAIL'
               else: 
                   d['attributes']['status'] = 'PASS'
        return d

    def genXMLdep(self, d,indent):
        s = ''
        tag = d['tag']
        if tag == 'dependencies':
            s += (2*indent)*" "
            s += '<dependencies>\n'
	elif tag in ['join','dependency'] :
            s += (2*indent)*" "
            s += '<%s'%(tag)
            for k in d['attributes'].keys():
              s += " %s='%s'"%(k,d['attributes'][k])
            if tag == 'dependency': s+= '/'
            s += '>\n' 
        for child in d['children']:
            s += self.genXMLdep(child,indent+1)
        if tag in ['dependencies','join']:
            s += (2*indent)*" "
            s += "</%s>\n"%(d['tag'])
        return s

    def pcheck(self, itest):
        ''' check the prerequisites '''
        rval = [1,'']
        if self.options.has_key('debug'): print 'D:\tpcheck() <%s>'%(itest.tagname)
        if itest.tagname != 'dependency':
            if hasattr(itest,'children'):
                rlist=[]
                # append to rval[1] (expression)
                if itest.tagname in ['OR','NOR','AND','NAND']: 
                    rval[1] += '('
                elif itest.tagname == 'NOT' and len(itest.children)>1:
                    raise TestError, '<NOT> must have only one child node'
                for i in itest.children:
                    tval = self.pcheck(i)
                    rlist.append(tval[0])
                    rval[0] = tval[0]
                    rval[1] += tval[1]
                    for j in tval[2:]: 
                        rval.append(j)
                    if itest.tagname == 'OR' or itest.tagname=='NOR':
                        if i != itest.children[len(itest.children)-1]: 
                            rval[1] += ' '+itest.tagname+' '
                    elif itest.tagname == 'AND' or itest.tagname=='NAND':
                        if i != itest.children[len(itest.children)-1]:
                            rval[1] += ' '+itest.tagname+' '
                    elif itest.tagname == 'NOT':
                        rval[1] = ' '+itest.tagname+' '+rval[1]
                if itest.tagname in ['OR','NOR','AND','NAND']: 
                    rval[1] += ')'
                if itest.tagname=='OR' or itest.tagname=='NOR':
                    rval[0] = 0
                    for i in rlist:
                        if i == 1:
                            rval[0] = 1
                    if itest.tagname=='NOR': 
                        rval[0] = not rval[0]
                elif itest.tagname=='AND' or itest.tagname=='NAND':
                    rval[0] = 1
                    for i in rlist:
                        if i == 0:
                            rval[0] = 0
                    if itest.tagname=='NAND': 
                        rval[0] = not rval[0]
                elif itest.tagname=='NOT':
                    rval[0] = not rlist[0]
        elif itest.tagname=='dependency':
            rval = [1, itest.name]
            parent = self.findTestByName(itest.name)
            if parent:				# wcm- for kludge to allow --test, --noDescendents, etc to work
               rval.append((parent.name,parent.pctMatched,itest.minPctMatch,itest.maxPctMatch))
               if parent.pctMatched < float(itest.minPctMatch) or \
                    parent.pctMatched > float(itest.maxPctMatch):
                   rval[0] = 0
        return rval


    def execute(self):
        '''TestList.execute() : Main driver for tests. '''

        # print column headers
        print ""
        print "iterations\ttest name\t% matched\tPass/Fail\tmessage"
        print "----------\t---------\t---------\t---------\t----------"

        sleep(float(self.preDelay))
     
        for listIter in range(int(self.numReps)):

            # Iterate over all tests in the sequence
            for test in self.tests:

                # update self.execData with test info
                self.filedata.addTest(test.name)
                self.filedata.updateTest(test.name, minPctMatch=float(test.minPctMatch), \
						    maxPctMatch=float(test.maxPctMatch), \
						    numReps=test.numReps, nMatch=test.nMatch, \
                                                    type=test.tagname)

                if self.options.has_key('verbose'): print "V:\tExecuting test '%s'"%(test.name)
                # Check the prerequisites
                try:
                    prereqs = self.pcheck(test)
                    d = self.pcheck2(test)
#                    print "============="
#                    print "---->", prereqs
#                    print "---->", d
                    s = self.genXMLdep(d,1)
                    self.filedata.updateTest(test.name,depXML=s)
                except TestError,msg:
                    print "Fatal error: %s"%(msg)
                    sys.exit(1)
                if prereqs[0] == 1:
                    try:
                        test.prepare(self.options, self.filedata.testdata[test.name])
                    except TestSetupError, e:
                        print "Test %s setup failed because of error: %s"%(test.name,e)
                    while test.irep < test.numReps:
                        try:
                            test.runTest(self.scratch)
                        except TestSetupError,e:
                            if self.options.has_key('verbose'):
                                print "V:\t[%d of %d] in %s failed execution :%s:"% \
                                        (test.irep,test.numReps,test.name,e)
                            if test.onMismatch == "BREAK": 
                                test.printSummary("BREAK")
                                break
                            elif test.onMismatch == "HALT": 
                                test.printSummary("HALT")
                                sys.exit(100)
                    if test.irep == test.numReps:
                        test.printSummary("")
                else:
                    print "[0 of %s]\t%-10s\tDEPENDENCY FAILURE(S)\t\t%s"%(test.numReps,test.name,prereqs[1])
                    for i in prereqs[2:]:
                        print "\t\t\t\t%s expected [%.1f%% : %.1f%%], got %s"%(i[0],float(i[2]),\
                              float(i[3]),i[1])
		self.filedata.updateTest(test.name, status=test.status, nMatch=test.nMatch, \
                                         pctMatched=test.pctMatched, failDeps=not prereqs[0])
            if listIter < int(self.numReps)-1:
                sleep(float(self.iterDelay)) 

        sleep(float(self.postDelay))

def printDict(D):
  for k in D.keys():
    print "%-10s"%k, "::", `D[k]`

# ------------------------------------------------------------------------------
# class TestSet
# ------------------------------------------------------------------------------
class TestSet(Node):
    __textfields__=['name']
    typemap={'info':TestInfo,'testList':TestList} 

    def __setdefaults__(self):
        self.filedata = filedata()

    def execute(self, options):
        ''' this is the file that is called by main() '''
        self.options = options
        # Execute the tests
        for child in self.children:
            if child.tagname=='testList':
                child.prepare(self.options, self.filedata)
                child.execute()
	#
	# generate result file
	#        self.filedata.title = self.name
	self.filedata.update(title=self.name, fname=self.options['filename'])
        for child in self.children:
            if child.tagname=='info':
                for cc in child.children:
                    self.filedata.addPerson(cc.tagname, cc.name, cc.phone, cc.email)
                if hasattr(child,'description'):
                    self.filedata.desc = strip(child.description)

	if self.options.has_key('logfile'):
	        OFP = open(self.options['logfile'],'a')
        	OFP.write(self.filedata.SummaryLine())
        	OFP.close()

	if self.options.has_key('resultfile'):
		OFP = open(self.options['resultfile'],'w')
		OFP.write(self.filedata.genXMLsummary())
		OFP.close()
                    

# EOF
